// Copyright (c) 2001, 2002 Zope Foundation and Contributors.
// All Rights Reserved.
//
// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE

package main
// Bits that should be in ZODB		XXX -> zodb

import (
	"context"

	"lab.nexedi.com/kirr/neo/go/zodb"

	pickle "github.com/kisielk/og-rek"
)

type Object struct {
	jar	*Connection
	oid	zodb.Oid
	serial	zodb.Tid
}

type PyObject struct {
	Object

	pyclass pickle.Class // python class of this object
	pystate interface{}  // object state. python passes this to pyclass.__new__().__setstate__()
}

// Connection represents a view of ZODB database.
//
// XXX Connection, and {Py}Object methods that relate to it, are not safe for
// modifications from multiple goroutines simultaneously.
type Connection struct {
	stor	zodb.IStorage	// underlying storage
	at	zodb.Tid	// current view of database
}

// Gets loads and decodes object from the database according to its current view.
//
// FIXME multiple calls to Get(oid) have to return the same instance.
// XXX   this is needed if there are several persistent references to the same object.
//       however wendelin.core does not do this.
func (conn *Connection) Get(ctx context.Context, oid zodb.Oid) (*PyObject, error) {
	// XXX -> loadpy
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return nil, err
	}

	pyclass, pystate, err := zodb.PyData(buf.Data).Decode()
	if err != nil {
		return nil, err	// XXX err ctx
	}

	buf.Release()

	return &PyObject{
		Object:  Object{jar: conn,oid: oid, serial: serial},
		pyclass: pyclass,
		pystate: pystate,
	}, nil
}

func (conn *Connection) loadpy(ctx context.Context, oid zodb.Oid) (pyclass pickle.Class, pystate interface{}, serial zodb.Tid, _ error) {
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return pickle.Class{}, nil, zodb.InvalidTID, err
	}

	pyclass, pystate, err = zodb.PyData(buf.Data).Decode()
	if err != nil {
		return pickle.Class{}, nil, zodb.InvalidTID, err	// XXX err ctx
	}

	buf.Release()

	return pyclass, pystate, serial, nil
}

// newGhost creates new ghost object.
func (conn *Connection) newGhost(pyclass pickle.Class, oid zodb.Oid) interface{} {
	pyobj := PyObject{
		Object:  Object{jar: conn, oid: oid, serial: zodb.InvalidTID},
		pyclass: pyclass,
		pystate: nil,
	}

	// TODO switch on pyclass and transform e.g. "zodb.BTree.Bucket" -> *ZBucket
	return &pyobj
}

// PDeactivates transforms object to ghost state.
//
// In ghost state object data is dropped and only oid/pyclass information is left in RAM.
//
// XXX if state=modified PDeactivate must be noop.
func (pyobj *PyObject) PDeactivate() {
	pyobj.pystate = nil
	pyobj.serial  = zodb.InvalidTID
}

// PActivate brings object to live state.
//
// If object state is not in RAM - it is loaded from the database.
func (pyobj *PyObject) PActivate(ctx context.Context) error {
	if pyobj.pystate != nil {
		return nil // already loaded
	}

	pyclass, pystate, serial, err := pyobj.jar.loadpy(ctx, pyobj.oid)
	if err != nil {
		return err
	}

	if pyclass != pyobj.pyclass {
		// XXX complain pyclass changed
		// XXX both ref and object data uses pyclass so it indeed can be different
	}

	pyobj.serial  = serial
	pyobj.pystate = pystate
	return nil
}
